/*

 eth.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "eth.h"
#include "crc8.h"

#define SERVER_IP INADDR_ANY
#define SERVER_PORT 64000

static int sock = -1;
static unsigned char ebufi[1030], ebufo[1030];
static unsigned int ebufip, ebufop, ebufil;
static struct sockaddr_in eserver;
static struct sockaddr_in eclient;
static socklen_t fromlen = sizeof(struct sockaddr_in);
static struct in_addr sin_addr;

static const char *drivername(void) {
    return "ETH";
}

static int initialize(const char *addr, int network, FILE *lf) {
    struct hostent *hp = addr ? gethostbyname(addr) : NULL;
    if (hp) memcpy(&sin_addr, hp->h_addr, sizeof(sin_addr));
    else sin_addr.s_addr = htonl(INADDR_ANY);

    if (sock >= 0) close(sock);

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        fprintf(lf, "Cannot open socket\n");
        return -1;
    }
    memset(&eserver, 0, sizeof(eserver));
    eserver.sin_family = AF_INET;
    eserver.sin_addr.s_addr = SERVER_IP;
    eserver.sin_port = htons(SERVER_PORT + network);
    if (bind(sock, (struct sockaddr *)&eserver, sizeof(eserver))<0) {
        fprintf(lf, "Cannot bind\n");
        close(sock);
        sock = -1;
        return -1;
    }
    return 0;
}

static int getb(void) {
    if (ebufip >= sizeof(ebufi)) return -EIO;
    if (ebufip >= ebufil) return -EIO;
    return ebufi[ebufip++];
}

static int sendb(unsigned char a) {
    if (ebufop >= sizeof(ebufo)) return -EIO;
    ebufo[ebufop++] = a;
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    if (ebufip + bytes > ebufil) {
        memcpy(data, ebufi + ebufip, ebufil - ebufip);
        ebufip = ebufil;
        return -EIO;
    }
    memcpy(data, ebufi + ebufip, bytes);
    ebufip += bytes;
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    if (ebufop + bytes > sizeof(ebufo)) {
        memcpy(ebufo + ebufop, data, sizeof(ebufo) - ebufip);
        ebufop = sizeof(ebufo);
        return -EIO;
    }
    memcpy(ebufo + ebufop, data, bytes);
    ebufop += bytes;
    return 0;
}

static void eshutdown(void) {
    if (sock >=0 ) close(sock);
    sock = -1;
    return;
}

static int reinitialize(FILE *UNUSED(lf)) {
    return 0;
}

static int flush(void) {
    if (!ebufop) return 0;
    if (sendto(sock, ebufo, ebufop, 0, (struct sockaddr *)&eclient, fromlen)<0) {
        ebufop = 0;
        return -EIO;
    }
    ebufop = 0;
    return 0;
}

static void clean(void) {
    return;
}

static int waitb(void) {
    int n;
    if (sock < 0) return -ENODEV;
    ebufip = ebufil = 0;
    n = recvfrom(sock, ebufi+2, sizeof(ebufi)-2, 0, (struct sockaddr *)&eclient, &fromlen);
    if (n < 0) return -EIO;
    ebufi[0]=ntohs(eclient.sin_port) >> 8;
    ebufi[1]=ntohs(eclient.sin_port);
    eclient.sin_addr.s_addr = sin_addr.s_addr;
    ebufip = ebufi[0] ? 0 : 2;
    ebufil = n + 2;
    return ebufi[ebufip++];
}

const struct io_ops_s eth_io_ops = {
    .drivername   = drivername,
    .initialize   = initialize,
    .reinitialize = reinitialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .wait         = waitb,
    .clean        = clean,
};
