/*

 rs232.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>

HANDLE commhandle; 

#else
#include <sys/ioctl.h>
#include <termios.h>

int commhandle;

#ifndef FNDELAY
#define FNDELAY 0
#endif

#endif

#include "rs232.h"
#include "crc8.h"

static const char *drivername(void) {
    return "RS232";
}

static const char *i_dev;
static int i_baudlow;

static int initialize(const char *dev, int baudlow, FILE *lf) {
#ifdef WIN32
    DCB dcb;
    COMMTIMEOUTS commtimeouts;
// Open the serial port
    commhandle=CreateFile(dev, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
// Initialize the DCBlength member 
    dcb.DCBlength = sizeof (DCB);
// Get the default port setting information
    GetCommState (commhandle, &dcb);
//Change the DCB structure settings
    dcb.BaudRate = baudlow ? 38400 : 115200;// Current baud
    dcb.fBinary = TRUE;               	// Binary mode, (windows supports only binary)
    dcb.fParity = TRUE;               	// Enable parity checking 
    dcb.fOutxCtsFlow = TRUE;         	// CTS output flow control 
    dcb.fOutxDsrFlow = FALSE;         	// No DSR output flow control 
    dcb.fDtrControl = DTR_CONTROL_ENABLE; 	// Enable DTR signal while device open
    dcb.fDsrSensitivity = FALSE;      	// DSR sensitivity 
    dcb.fTXContinueOnXoff = TRUE;     	// XOFF continues Tx 
    dcb.fOutX = FALSE;                	// No XON/XOFF out flow control 
    dcb.fInX = FALSE;                 	// No XON/XOFF in flow control 
    dcb.fErrorChar = FALSE;           	// Disable error replacement 
    dcb.fNull = FALSE;                	// Disable null stripping 
    dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;// RTS flow control 
    dcb.fAbortOnError = FALSE;        	// Do not abort reads/writes on error
    dcb.ByteSize = 8;                 	// Number of bits/byte, 4-8 
    dcb.Parity = NOPARITY;            	// 0-4=no,odd,even,mark,space 
    dcb.StopBits = ONESTOPBIT;        	// 0,1,2 = 1, 1.5, 2 
// Configure the port according to the specifications of the DCB structure
    if (!SetCommState (commhandle, &dcb)) {
            fprintf(lf, "Unable to configure the serial port\n");
            return -ENODEV;
    }
// Retrieve the timeout parameters for all read and write operations on the port 
    GetCommTimeouts (commhandle, &commtimeouts);

// Change the COMMTIMEOUTS structure settings
    commtimeouts.ReadIntervalTimeout = MAXDWORD;  
    commtimeouts.ReadTotalTimeoutMultiplier = 0;  
    commtimeouts.ReadTotalTimeoutConstant = 1000;	// MAXDWORD, 0, after x miliseconds read returns even if no bytes received
    commtimeouts.WriteTotalTimeoutMultiplier = 0;  
    commtimeouts.WriteTotalTimeoutConstant = 1000;

// Set the timeout parameters for all read and write operations on the port 
    if (!SetCommTimeouts (commhandle, &commtimeouts)) {
            fprintf(lf, "Unable to set the timeout parameters\n");
            return -ENODEV;
    }
#else
    int flags;
    struct termios options;

    /// Changes by Silver Dream !
    if ((commhandle = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0)
    {
        fprintf(lf, "Cannot open %s\n",dev);
        return -ENODEV;
    }

    if (ioctl(commhandle, TIOCEXCL) < 0)
    {
        fprintf(lf, "Error setting TIOCEXCL on %s - %s(%d).\n", dev, strerror(errno), errno);
        return -ENODEV;
    }

    if (fcntl(commhandle, F_SETFL, 0) < 0)
    {
        fprintf(lf, "Error clearing O_NONBLOCK %s - %s(%d).\n", dev, strerror(errno), errno);
        return -ENODEV;
    }

    if (tcgetattr(commhandle, &options) < 0)
    {
        fprintf(lf, "Cannot read params %s\n", dev);
        return -ENODEV;
    }
    // End of changes by Silver Dream !

    /* baud rate */

    if (baudlow) {
        cfsetispeed(&options, B38400);
        cfsetospeed(&options, B38400);
    } else {
        cfsetispeed(&options, B115200);
        cfsetospeed(&options, B115200);
    }

    /* input options */

    options.c_iflag = 0;

    /* cout options */

    options.c_oflag = 0;

    /* control options */
    /* 8N1 */

    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~CSTOPB;
    options.c_cflag &= ~CSIZE;
    options.c_cflag |= CS8;

    /* CTS/RTS flow control */

//    if (flowctrl)
//    {
//	printf("Enabling RTS/CTS flow control\n");
    options.c_cflag |= CRTSCTS;
//    } else options.c_cflag &= ~CRTSCTS;

    /* line options */
    /* raw input */
 
//  options.c_lflag |= IEXTEN; /* leave enabled to get flow control */
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ISIG);

    /* control character */

//  options.c_cc[VMIN] = 1; /* minimum number of character to read */
    options.c_cc[VMIN] = 0; /* minimum number of character to read */
    options.c_cc[VTIME] = 10; /* time to wait for data */

    if (tcsetattr(commhandle, TCSANOW, &options) < 0)
    {
        fprintf(lf, "Cannot set params on %s\n", dev);
        return -ENODEV;
    }

  /* turn turn O_NONBLOCK again for sure */
  /* O_NONBLOCK = O_NDELAY -> DOESNT CARE WHAT STATE DCD */


    flags = fcntl(commhandle, F_GETFL, 0);
    if (flags < 0)
    {
        fprintf(lf, "Error executing fcntl on %s\n", dev);
        return -ENODEV;
    }
    if (fcntl(commhandle, F_SETFL, flags & ~(O_NONBLOCK | FNDELAY) ) < 0)
    {
        fprintf(lf, "Error executing fcntl on %s\n", dev);
        return -ENODEV;
    }

//fcntl(commhandle, F_SETFL, (flags | O_NONBLOCK) & ~FNDELAY);
  /* set the read function to return immediately */

//  fcntl(commhandle, F_SETFL, FNDELAY);
#endif
    i_dev = dev; i_baudlow = baudlow;
    return 0;
}

static int getb() {
    unsigned char data;
    int err;

#ifndef WIN32
    err=read(commhandle, &data, 1);
    if (err < 0) return -EIO;
    if (err == 1) {
        crc_add(data);
        return (unsigned char)data;
    }
#else
    DWORD byteswritten;

    err=ReadFile(commhandle, &data, 1, &byteswritten, NULL);
    if (err) {
        crc_add(data);
        return (unsigned char)data;
    }
#endif
    return -EBUSY;
}

static int sendb(unsigned char a) {
    int err;

#ifndef WIN32
    crc_add(a);
    do {
        err=write(commhandle, &a, 1);
        if (err < 0 ) return -EIO;
    } while (err != 1);
#else
    DWORD byteswritten;
    crc_add(a);
    err=WriteFile(commhandle, &a, 1, &byteswritten, NULL);
    if (!err) return -EIO;
#endif
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int err;

    do {
        int l;
#ifndef WIN32
        err=read(commhandle, data, bytes);
        if (err <= 0 ) return -EIO;
	for (l=0;l < err;l++) crc_add(data[l]);
	bytes-=err;data+=err;
#else
    DWORD dwBytesRead;
        err=ReadFile(commhandle, data, bytes, &dwBytesRead, NULL);
        if (!err || !dwBytesRead) return -EIO;
        for (l=0;l < dwBytesRead;l++) crc_add(data[l]);
	bytes-=dwBytesRead;data+=dwBytesRead;
#endif
    } while (bytes > 0);
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    int err;

    for (err=0;err < bytes;err++) crc_add(data[err]);
    do {
#ifndef WIN32
	err=write(commhandle, data, bytes);
        if (err < 0 ) return -EIO;
	bytes-=err;data+=err;
#else
        DWORD dwBytesWritten;
	err=WriteFile(commhandle, data, bytes, &dwBytesWritten, NULL);
        if (!err || !dwBytesWritten) return -EIO;
	bytes-=dwBytesWritten;data+=dwBytesWritten;
#endif
    } while (bytes > 0);
    return 0;
}

static void shutdown(void) {
#ifndef WIN32
    tcflush(commhandle,TCIOFLUSH);
    close(commhandle);
#else
    CancelIo(commhandle);
    CloseHandle(commhandle);
#endif
    return;
}

static int reinitialize(FILE *lf) {
    shutdown();
    return initialize(i_dev, i_baudlow, lf);
}

static int flush(void) {
    int i;
#ifndef WIN32
    alarm(1);
    i=tcdrain(commhandle);
    alarm(0);
#else
    i=!FlushFileBuffers(commhandle);
#endif
    return i;
}

static void clean(void) {
#ifndef WIN32
    tcflush(commhandle,TCIOFLUSH);
#else
    CancelIo(commhandle);
#endif
    return;
}

static int waitb(void) {
    int b;
    for (;;) {
	b=getb();
        if (b != -EBUSY) return b;
    }
}

const struct io_ops_s rs232_io_ops = {
    .drivername   = drivername,
    .initialize   = initialize,
    .reinitialize = reinitialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = shutdown,
    .flush        = flush,
    .wait         = waitb,
    .clean        = clean,
};
